"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createWeakStorage = exports.breakdownArgs = void 0;
var tslib_1 = require("tslib");
var utils_1 = require("./utils");
exports.breakdownArgs = function (args) {
    var weaks = [];
    var strongs = [];
    for (var i = 0; i < args.length; i++) {
        if (utils_1.isWeakable(args[i])) {
            weaks.push(args[i]);
        }
        else {
            strongs.push(args[i]);
        }
    }
    return tslib_1.__spreadArrays(weaks, strongs);
};
exports.createWeakStorage = function (storage) {
    if (storage === void 0) { storage = new WeakMap(); }
    return ({
        get: function (args) {
            var slices = exports.breakdownArgs(args);
            if (!slices.length) {
                throw new Error("No weak-mappable (object, function, symbol) argument found.");
            }
            var readFrom = storage;
            var test = { weak: storage };
            for (var i = 0; i < slices.length; ++i) {
                var storageKey = utils_1.isWeakable(slices[i]) ? 'weak' : 'strong';
                readFrom = test[storageKey];
                test = readFrom.get(slices[i]);
                if (!test) {
                    // get: no forward
                    return undefined;
                }
            }
            if (!readFrom) {
                return undefined;
            }
            return {
                value: test.storedValue,
            };
        },
        set: function (args, value) {
            var _a;
            var slices = exports.breakdownArgs(args);
            var writeTo = storage;
            var next = { weak: storage };
            for (var i = 0; i < slices.length; ++i) {
                var _b = utils_1.isWeakable(slices[i]) ? ['weak', function () { return new WeakMap(); }] : ['strong', function () { return new Map(); }], storageKey = _b[0], factory = _b[1];
                writeTo = next[storageKey];
                if (!writeTo) {
                    next[storageKey] = writeTo = factory();
                }
                next = writeTo.get(slices[i]);
                if (!next || !next[storageKey]) {
                    next = (_a = {},
                        _a[storageKey] = factory(),
                        _a);
                    writeTo.set(slices[i], next);
                }
            }
            next.storedValue = value;
            return value;
        },
    });
};
